﻿// Learn more about F# at http://fsharp.net

///////////////
// Initial object algebra interface for expressions: integers and addition
type ExpAlg<'E> =
  abstract lit : int -> 'E
  abstract add : 'E -> 'E -> 'E

///////////////
// An object algebra implementing that interface (evaluation)

// The evaluation interface
type Eval = unit -> int

// The object algebra
type EvalExpAlg () =
  interface ExpAlg<Eval> with
    member this.lit x = fun () -> x
    member this.add e1 e2 = fun () -> e1() + (e2())

///////////////
// Evolution 1: Adding subtraction
type SubExpAlg<'E> =
  inherit ExpAlg<'E> 
  abstract sub : 'E -> 'E -> 'E

// Updating evaluation:
type EvalSubExpAlg () =
  inherit EvalExpAlg () 
  interface SubExpAlg<Eval> with
    member this.sub e1 e2 = fun () -> e1() - (e2())

///////////////
// Evolution 2: Adding pretty printing
type PPrint = unit -> string

type PrintExpAlg ()  =
  interface SubExpAlg<PPrint> with
    member this.lit n = fun () -> n.ToString()
    member this.add e1 e2 = fun () -> (e1 ()) + " + " + (e2 ())
    member this.sub e1 e2 = fun () -> (e1 ()) + " - " + (e2 ())

///////////////
// An alternative object algebra for pretty printing:

// Often, when precise control over the invocation of 
// methods is not needed, we can simplify object algebras. 
// For example, here's an alternative implementation 
// of pretty printing that directly computes a string:
type PrintExpAlg2 ()= 
  interface SubExpAlg<string> with
    member this.lit n = n.ToString()
    member this.add e1 e2 = e1 + " + " + e2 
    member this.sub e1 e2 = e1 + " - " + e2

///////////////
// Testing
let _ = 
  let exp1 (alg : ExpAlg<'E>) = alg.add (alg.lit 3) (alg.lit 4)
  let exp2 (alg : SubExpAlg<'E>) = alg.sub (exp1 alg) (alg.lit 4)
  let ea = new EvalExpAlg()
  let esa = new EvalSubExpAlg()
  let pa = new PrintExpAlg()
  let pa2 = new PrintExpAlg2()

  // We can call esa with exp1
  let ev = exp1(esa)

  // But calling ea with exp2 is an error
  //let ev_bad = exp2(ea)

  printfn "Evaluation of exp1 \"%s\" is: %d" (exp1 pa ()) (ev ())
  printfn "Evaluation of exp2 \"%s\" is: %d" (exp2 pa ()) (exp2 esa ())
  printfn "The alternative pretty printer works nicely too!\nexp1: %s\nexp2: %s" (exp1 pa2) (exp2 pa2)